#!/bin/bash

verbose_log=/dev/null  
progress_log=/dev/stderr

interface="external"  # pipe with redis-cli
limit=0        # < 1 then no limit

pause=0        # 0: no, 1: pause
interactive=("no" "yes")

flushdb=0      # 0:Don't Flash Redis DB

ejdic='ejdic-hand-txt/ejdic-hand-utf8.txt' #ejdic='e.txt'
gene='gene95/gene.txt' #gene='g.txt'

usage_exit() {
    echo "$0 -fpvPi -e [ejdict] -g [gene] -L [progress_log] -l [limit]"
    echo
    echo "Options:"
    echo "  -f Redisのデータベースをクリアする"
    echo "  -p 対話的に確認しながら実行"
    echo "  -v 冗長メッセージ表示(デバッグ用)"
    echo "  -i redis-cliをスクリプト内で呼び出す"
    echo "  -P redis-cli --pipe で大量データをインサートするプロトコル"
    echo "  -l [limit] それぞれの辞書でインポートするレコード数の上限を[limit]とする"
    echo "  -L [progress_log]  進捗ログメッセージの出力先を指定"
    echo "  -e [ejdict], -j [gene] 各辞書ファイルの場所を指定"
    echo
    echo "Examples:"
    echo "  $0 -f | redis-cli   # redis-cliとパイプで接続、最も基本的な使い方"
    echo "  $0 -fi              # 1レコードごとにredis-cliをスクリプト内から実行" 
    echo
    echo "  $0 -fP -L /dev/null > \"protocolfile\"  # 大量データインサートを高速に実施する"
    echo "  redis-cli --pipe < \"protocolfile\""
    echo
    echo "    いずれも、-fオプションで、既存のRedisデータをクリア"
    exit
}

while getopts fpvPie:g:L:l: OPT
do
  case $OPT in
    f) flushdb=1; ;;
    p) pause=1; ;;
    v) verbose_log="/dev/stderr"; ;;
    P) interface="pipe"; ;;        # redis-cli --pipe を使う場合
    i) interface="domestic"; ;;    # redis-cliをスクリプト内部で呼び出し
    e) ejdic=${OPTARG}; ;;
    g) gene=${OPTARG}; ;;
    L) progress_log=${OPTARG}; ;;
    l) limit=$(expr ${OPTARG} + 0) > /dev/null 2>&1
        if [ $? -gt 0 ]; then
            limint=0
        fi
        ;;
    \?) usage_exit ;;
  esac
done
shift $((OPTIND - 1))

if [ ! -f $ejdic ]; then
    echo "no $ejdic exist." >&2
    exit 1
elif [ ! -f $gene ]; then
    echo "no $gene exist." >&2
    exit 1
elif [ ! $(which redis-cli) ]; then
    echo "no redis-cli executable." >&2
    exit 1
fi

pause_n_query() {
    [ ${pause} -eq 0 ] && return
    read -p "$1" q >&2
    case $q in
        [nN]* ) echo "quit" >&2
                exit 1
        ;;
    esac
}

redis_sadd() { # bodyもheadも'があれば\'に置き換える
    local key=$(sed "s/'/\\\'/g" <<<$1)
    local values=$(sed "s/'/\\\'/g" <<<$2)

    printf "%0.7d:%s " "$3" "$key" > ${progress_log}
    case ${interface} in
      "external" ) # 直接リダイレクトする場合
        echo "sadd '${key}' '${values}'" ;;
      "pipe") # パイプラインで処理する場合
        printf "sadd '%s' '%s'\r\n" "${key}" "${values}" ;;
      "domestic") # 1行ごとにredis-cliを実行するので、すごく低速。$?を確認して、1行の処理ごとにエラー処理を書き加えることができる。
        redis-cli sadd "${key}" "${values}" ;;
    esac
    #sleep 0.2
}

do_import() {
    local dic=$1

    pause_n_query "Import ${dic} into Redis? (Y/n) "
    echo "Process ${dic}" > ${verbose_log}

    i=0
    case ${dic} in
      ${ejdic} )
        set -f
        while IFS= read -r line || [ -n "${line}" ] ; do
            head=$(cut -f 1 <<<${line})
            body=$(cut -f 2 <<<${line})
            # headに,があれば、分割 ... 分割した数だけ、bodyとセットにして登録
            ifs=${IFS}
            IFS=","
            set - ${head}
            while [ ! -z ${1} ] ; do
                let i++
                echo "${i}: ${1}:${body}" > ${verbose_log}
                redis_sadd "${1}" "${body}" "${i}"
                shift
            done
            IFS=${ifs}
            [ ${limit} -gt 0 -a ${i} -ge ${limit} ] && break
        done < ${dic}
        ;;

      ${gene} )
        set +f
        while IFS= read -r line || [ -n "${line}" ] ; do
            if (( i > 1 )) ; then  # 最初の2行はスキップ
                if (( i % 2 == 0 )) ; then
                    head=${line}
                else
                    body=${line}
                    echo "${i}: ${head}:${body}" > ${verbose_log}
                    redis_sadd "${head}" "${body}" "${i}"
                fi
            fi
            let i++
            [ ${limit} -gt 0 -a $(((i-2)/2)) -ge ${limit} ] && let i=(i-2)/2 && break
        done < ${gene}
        ;;

      * )
        echo "Wrong files ${dic}." > ${verbose_log}
        ;;
    esac

    echo "${dic} done, ${i} items" > ${progress_log}

    echo > ${verbose_log}
    echo > ${verbose_log}
}

echo "${0}" >&2
echo " Flash Redis DB=${interactive[$flushdb]}" >&2
echo " Import Interface=${interface}" >&2
echo " verbose_log=${verbose_log}" >&2
echo " progress_log=${progress_log}" >&2
echo " limit=${limit}" >&2
echo " Interactive=${interactive[$pause]}" >&2
echo >&2

if [ ${flushdb} -ne 0 ]; then
    pause_n_query "Flashing Redis DB, continue? (Y/n) "
    echo "flushall" >&2
    case $interface in
      "external" ) echo "flushall" ;;
      "pipe" ) printf "flushall\r\n" ;;
      "domestic") redis-cli flushall &> /dev/null ;;
      *) ;;
    esac
fi

for d in ${ejdic} ${gene} ; do
    do_import $d
done
